@@ -4,6 +4,7 @@ package importpaths
 
 import (
 	"errors"
+	"fmt"
 	"go/parser"
 	"go/printer"
 	"go/token"
@@ -37,8 +38,9 @@ func Rewrite(dir string, replace ReplaceFunc) error {
 			if name == dir {
 				return nil
 			}
-			// don't recurse into vendor
-			if info.Name() == "vendor" {
+			// don't recurse into vendor, testdata, or node_modules directories
+			switch info.Name() {
+			case "vendor", "testdata", "node_modules":
 				return filepath.SkipDir
 			}
 			// don't recurse into underscore-prefix or dot-prefix directories
@@ -74,53 +76,50 @@ func RewriteFile(name string, replace ReplaceFunc) error {
 	// if we need to write it back out.
 	f, err := parser.ParseFile(fset, name, nil, parser.ParseComments)
 	if err != nil {
-		e := err.Error()
-		msg := "expected 'package', found 'EOF'"
-		if e[len(e)-len(msg):] == msg {
-			return nil
-		}
 		return err
 	}
 	// iterate through the import paths. if a change occurs update bool.
 	change := false
 	for _, i := range f.Imports {
+		pos := fset.Position(i.Pos())
 		// unquote the import path value.
 		path, err := strconv.Unquote(i.Path.Value)
 		if err != nil {
-			return err
+			return fmt.Errorf("%s: %w", pos, err)
 		}
 		// replace the value using the replace function
-		path, err = replace(fset.Position(i.Pos()), path)
+		path, err = replace(pos, path)
 		if err != nil {
 			if err == ErrSkip {
 				continue
 			}
-			return err
+			return fmt.Errorf("%s: %w", pos, err)
 		}
 		i.Path.Value = strconv.Quote(path)
 		change = true
 	}
+	pkgpos := fset.Position(f.Package)
 	for _, cg := range f.Comments {
 		for _, c := range cg.List {
-			if strings.HasPrefix(c.Text, "// import \"") {
+			pos := fset.Position(c.Pos())
+			const prefix = "// import "
+			if pos.Line == pkgpos.Line && strings.HasPrefix(c.Text, prefix) {
 				// trim off extra comment stuff
-				ctext := c.Text
-				ctext = strings.TrimPrefix(ctext, "// import")
-				ctext = strings.TrimSpace(ctext)
+				ctext := strings.TrimSpace(c.Text[len(prefix):])
 				// unquote the comment import path value
 				ctext, err := strconv.Unquote(ctext)
 				if err != nil {
-					return err
+					return fmt.Errorf("%s: %v", pos, err)
 				}
 				// match the comment import path with the given replacement map
-				ctext, err = replace(fset.Position(c.Pos()), ctext)
+				ctext, err = replace(pos, ctext)
 				if err != nil {
 					if err == ErrSkip {
 						continue
 					}
-					return err
+					return fmt.Errorf("%s: %w", pos, err)
 				}
-				c.Text = "// import " + strconv.Quote(ctext)
+				c.Text = prefix + strconv.Quote(ctext)
 				change = true
 			}
 		}
@@ -166,7 +165,7 @@ type RewriteModuleOptions struct {
 	NewVersion string
 	NewPrefix  string
 	PkgDir     string
-	OnRewrite  func(pos token.Position, oldpath, newpath string)
+	OnRewrite  func(pos token.Position, oldpath, newpath string) error
 }
 
 // RewriteModule rewrites imports of a specific module to a new version or prefix.
@@ -189,7 +188,9 @@ func RewriteModule(dir string, opt RewriteModuleOptions) error {
 			return "", ErrSkip
 		}
 		if opt.OnRewrite != nil {
-			opt.OnRewrite(pos, path, newpath)
+			if err := opt.OnRewrite(pos, path, newpath); err != nil {
+				return "", err
+			}
 		}
 		return newpath, nil
 	})
