@@ -2,28 +2,123 @@ package modproxy
 
 import (
 	"bufio"
-	"bytes"
+	"encoding/json"
+	"errors"
 	"fmt"
-	"io/ioutil"
+	"io"
 	"net/http"
+	"net/url"
 	"os"
 	"path"
+	"path/filepath"
+	"slices"
 	"strconv"
 	"strings"
 
+	"golang.org/x/mod/modfile"
 	"golang.org/x/mod/module"
 	"golang.org/x/mod/semver"
 	"golang.org/x/sync/errgroup"
 
+	"github.com/icholy/gomajor/internal/goenv"
 	"github.com/icholy/gomajor/internal/packages"
 )
 
+// Request sends requests to the module proxies in order and returns
+// the first 200 response.
+func Request(path string, cached bool) (*http.Response, error) {
+	proxies := goenv.GOPROXYURL()
+	if len(proxies) == 0 {
+		return nil, errors.New("no GOPROXY urls available")
+	}
+	var last *http.Response
+	for _, proxy := range proxies {
+		res, err := doProxyRequest(proxy, path, cached)
+		if err != nil {
+			return nil, err
+		}
+		if res.StatusCode == http.StatusOK {
+			return res, nil
+		}
+		last = res
+	}
+	return last, nil
+}
+
+func doProxyRequest(proxy *url.URL, path string, cached bool) (*http.Response, error) {
+	switch proxy.Scheme {
+	case "http", "https":
+		return httpRequest(proxy, path, cached)
+	case "file":
+		return fileRequest(proxy, path)
+	default:
+		return nil, errors.New("unsupported protocol " + proxy.Scheme)
+	}
+}
+
+func httpRequest(proxy *url.URL, path string, cached bool) (*http.Response, error) {
+	req, err := http.NewRequest(http.MethodGet, proxy.JoinPath(path).String(), nil)
+	if err != nil {
+		return nil, err
+	}
+	req.Header.Set("User-Agent", "GoMajor/1.0")
+	if cached {
+		req.Header.Set("Disable-Module-Fetch", "true")
+	}
+	return http.DefaultClient.Do(req)
+}
+
+func fileRequest(proxy *url.URL, path string) (*http.Response, error) {
+	root := proxy.Path
+	if filepath.VolumeName(root) != "" {
+		root, _ = strings.CutPrefix(root, "/")
+	}
+	f, err := os.Open(filepath.Join(
+		filepath.FromSlash(root),
+		filepath.FromSlash(path),
+	))
+	if err != nil {
+		if os.IsNotExist(err) {
+			return &http.Response{
+				StatusCode: http.StatusNotFound,
+				Status:     http.StatusText(http.StatusNotFound),
+				Body:       http.NoBody,
+			}, nil
+		}
+		return nil, err
+	}
+	fi, err := f.Stat()
+	if err != nil {
+		f.Close()
+		return nil, err
+	}
+	return &http.Response{
+		StatusCode:    http.StatusOK,
+		Status:        http.StatusText(http.StatusOK),
+		Body:          f,
+		ContentLength: fi.Size(),
+	}, nil
+}
+
 // Module contains the module path and versions
 type Module struct {
 	Path     string
 	Versions []string
 }
 
+// MaxVersion returns the latest version of the module in the list.
+// If pre is false, pre-release versions will are excluded.
+// Retracted versions are excluded.
+func MaxVersion(mods []*Module, pre bool, r Retractions) (*Module, string) {
+	for i := len(mods); i > 0; i-- {
+		mod := mods[i-1].Retract(r)
+		if max := mod.MaxVersion("", pre); max != "" {
+			return mod, max
+		}
+	}
+	return nil, ""
+}
+
 // MaxVersion returns the latest version.
 // If there are no versions, the empty string is returned.
 // Prefix can be used to filter the versions based on a prefix.
@@ -37,11 +132,22 @@ func (m *Module) MaxVersion(prefix string, pre bool) string {
 		if !pre && semver.Prerelease(v) != "" {
 			continue
 		}
-		max = MaxVersion(v, max)
+		if CompareVersion(max, v) < 0 {
+			max = v
+		}
 	}
 	return max
 }
 
+// Retract returns a copy of m with the retracted versions removed.
+func (m *Module) Retract(r Retractions) *Module {
+	versions := slices.Clone(m.Versions)
+	return &Module{
+		Path:     m.Path,
+		Versions: slices.DeleteFunc(versions, r.Includes),
+	}
+}
+
 // IsNewerVersion returns true if newversion is greater than oldversion in terms of semver.
 // If major is true, then newversion must be a major version ahead of oldversion to be considered newer.
 func IsNewerVersion(oldversion, newversion string, major bool) bool {
@@ -51,37 +157,34 @@ func IsNewerVersion(oldversion, newversion string, major bool) bool {
 	return semver.Compare(oldversion, newversion) < 0
 }
 
-// MaxVersion returns the larger of two versions according to semantic version precedence.
+// CompareVersion returns -1 if v < w, 1 if v > w, and 0 if v == w
 // Incompatible versions are considered lower than non-incompatible ones.
 // Invalid versions are considered lower than valid ones.
 // If both versions are invalid, the empty string is returned.
-func MaxVersion(v, w string) string {
+func CompareVersion(v, w string) int {
 	// sort by validity
 	vValid := semver.IsValid(v)
 	wValid := semver.IsValid(w)
 	if !vValid && !wValid {
-		return ""
+		return 0
 	}
 	if vValid != wValid {
 		if vValid {
-			return v
+			return 1
 		}
-		return w
+		return -1
 	}
 	// sort by compatibility
 	vIncompatible := strings.HasSuffix(semver.Build(v), "+incompatible")
 	wIncompatible := strings.HasSuffix(semver.Build(w), "+incompatible")
 	if vIncompatible != wIncompatible {
 		if wIncompatible {
-			return v
+			return 1
 		}
-		return w
+		return -1
 	}
 	// sort by semver
-	if semver.Compare(v, w) == 1 {
-		return v
-	}
-	return w
+	return semver.Compare(v, w)
 }
 
 // NextMajor returns the next major version after the provided version
@@ -124,23 +227,14 @@ func Query(modpath string, cached bool) (*Module, bool, error) {
 	if err != nil {
 		return nil, false, err
 	}
-	url := fmt.Sprintf("https://proxy.golang.org/%s/@v/list", escaped)
-	req, err := http.NewRequest(http.MethodGet, url, nil)
-	if err != nil {
-		return nil, false, err
-	}
-	req.Header.Set("User-Agent", "GoMajor/1.0")
-	if cached {
-		req.Header.Set("Disable-Module-Fetch", "true")
-	}
-	res, err := http.DefaultClient.Do(req)
+	res, err := Request(path.Join(escaped, "@v", "list"), cached)
 	if err != nil {
 		return nil, false, err
 	}
 	defer res.Body.Close()
 	if res.StatusCode != http.StatusOK {
-		body, _ := ioutil.ReadAll(res.Body)
-		if res.StatusCode == http.StatusNotFound && bytes.HasPrefix(body, []byte("not found:")) {
+		body, _ := io.ReadAll(res.Body)
+		if res.StatusCode == http.StatusNotFound {
 			return nil, false, nil
 		}
 		msg := string(body)
@@ -161,9 +255,36 @@ func Query(modpath string, cached bool) (*Module, bool, error) {
 	return &mod, true, nil
 }
 
+// ErrNoVersions is returned when the proxy has no version for a module
+var ErrNoVersions = errors.New("no module versions found")
+
 // Latest finds the latest major version of a module
 // cached sets the Disable-Module-Fetch: true header
-func Latest(modpath string, cached bool) (*Module, error) {
+// pre controls whether to return modules which only contain pre-release versions.
+func Latest(modpath string, cached, pre bool) (*Module, error) {
+	mods, err := List(modpath, cached)
+	if err != nil {
+		return nil, err
+	}
+	// find the retractions
+	var r Retractions
+	if mod, _ := MaxVersion(mods, false, nil); mod != nil {
+		var err error
+		r, err = FetchRetractions(mod)
+		if err != nil {
+			return nil, err
+		}
+	}
+	mod, _ := MaxVersion(mods, pre, r)
+	if mod == nil {
+		return nil, ErrNoVersions
+	}
+	return mod, nil
+}
+
+// List finds all the major versions of a module
+// cached sets the Disable-Module-Fetch: true header
+func List(modpath string, cached bool) ([]*Module, error) {
 	latest, ok, err := Query(modpath, cached)
 	if err != nil {
 		return nil, err
@@ -171,10 +292,11 @@ func Latest(modpath string, cached bool) (*Module, error) {
 	if !ok {
 		return nil, fmt.Errorf("module not found: %s", modpath)
 	}
+	history := []*Module{latest}
 	for i := 0; i < 100; i++ {
 		nextpath, ok := latest.NextMajorPath()
 		if !ok {
-			return latest, nil
+			return history, nil
 		}
 		next, ok, err := Query(nextpath, cached)
 		if err != nil {
@@ -195,9 +317,10 @@ func Latest(modpath string, cached bool) (*Module, error) {
 			}
 		}
 		if !ok {
-			return latest, nil
+			return history, nil
 		}
 		latest = next
+		history = append(history, latest)
 	}
 	return nil, fmt.Errorf("request limit exceeded")
 }
@@ -236,12 +359,89 @@ func QueryPackage(pkgpath string, cached bool) (*Module, error) {
 	return nil, fmt.Errorf("failed to find module for package: %s", pkgpath)
 }
 
+// FetchRetractions fetches the retractions for this module.
+func FetchRetractions(mod *Module) (Retractions, error) {
+	max := mod.MaxVersion("", false)
+	if max == "" {
+		return nil, nil
+	}
+	escaped, err := module.EscapePath(mod.Path)
+	if err != nil {
+		return nil, err
+	}
+	res, err := Request(path.Join(escaped, "@v", max+".mod"), false)
+	if err != nil {
+		return nil, err
+	}
+	defer res.Body.Close()
+	body, err := io.ReadAll(res.Body)
+	if err != nil {
+		return nil, err
+	}
+	if res.StatusCode != http.StatusOK {
+		msg := string(body)
+		if msg == "" {
+			msg = res.Status
+		}
+		return nil, fmt.Errorf("proxy: %s", msg)
+	}
+	file, err := modfile.ParseLax(mod.Path, body, nil)
+	if err != nil {
+		return nil, err
+	}
+	var retractions Retractions
+	for _, r := range file.Retract {
+		retractions = append(retractions, VersionRange{Low: r.Low, High: r.High})
+	}
+	return retractions, nil
+}
+
+// VersionRange is an inclusive version range.
+type VersionRange struct {
+	Low, High string
+}
+
+// Includes reports whether v is in the inclusive range
+func (r VersionRange) Includes(v string) bool {
+	return CompareVersion(v, r.Low) >= 0 && CompareVersion(v, r.High) <= 0
+}
+
+// Retractions is a list of retracted versions.
+type Retractions []VersionRange
+
+// Includes reports whether v is retracted
+func (rr Retractions) Includes(v string) bool {
+	for _, r := range rr {
+		if r.Includes(v) {
+			return true
+		}
+	}
+	return false
+}
+
 // Update reports a newer version of a module.
 // The Err field will be set if an error occured.
 type Update struct {
-	Module  module.Version
-	Version string
-	Err     error
+	Module module.Version
+	Latest module.Version
+	Err    error
+}
+
+// MarshalJSON implements json.Marshaler
+func (u Update) MarshalJSON() ([]byte, error) {
+	var err string
+	if u.Err != nil {
+		err = u.Err.Error()
+	}
+	return json.Marshal(struct {
+		Module module.Version
+		Latest module.Version
+		Err    string `json:",omitempty"`
+	}{
+		Module: u.Module,
+		Latest: u.Latest,
+		Err:    err,
+	})
 }
 
 // UpdateOptions specifies a set of modules to check for updates.
@@ -259,7 +459,7 @@ func Updates(opt UpdateOptions) {
 	ch := make(chan Update)
 	go func() {
 		defer close(ch)
-		private := os.Getenv("GOPRIVATE")
+		private := goenv.Get("GOPRIVATE")
 		var group errgroup.Group
 		if opt.Cached {
 			group.SetLimit(3)
@@ -272,14 +472,23 @@ func Updates(opt UpdateOptions) {
 				continue
 			}
 			group.Go(func() error {
-				mod, err := Latest(m.Path, opt.Cached)
+				mod, err := Latest(m.Path, opt.Cached, opt.Pre)
+				if err == ErrNoVersions {
+					return nil
+				}
 				if err != nil {
 					ch <- Update{Module: m, Err: err}
 					return nil
 				}
 				v := mod.MaxVersion("", opt.Pre)
 				if IsNewerVersion(m.Version, v, opt.Major) {
-					ch <- Update{Module: m, Version: v}
+					ch <- Update{
+						Module: m,
+						Latest: module.Version{
+							Path:    mod.WithMajorPath(v),
+							Version: v,
+						},
+					}
 				}
 				return nil
 			})
