@@ -1,69 +1,13 @@
 package modproxy
 
 import (
-	"fmt"
+	"net/http/httptest"
+	"reflect"
 	"testing"
+
+	"github.com/icholy/gomajor/internal/modproxy/testmodproxy"
 )
 
-func TestLatest(t *testing.T) {
-	tests := []string{
-		"github.com/go-redis/redis",
-		"github.com/russross/blackfriday",
-	}
-	for _, tt := range tests {
-		t.Run(tt, func(t *testing.T) {
-			mod, err := Latest(tt, true)
-			if err != nil {
-				t.Fatal(err)
-			}
-			t.Logf("Latest %s %s", mod.Path, mod.MaxVersion("", false))
-		})
-	}
-}
-
-func TestQuery(t *testing.T) {
-	mod, ok, err := Query("github.com/DATA-DOG/go-sqlmock", true)
-	if err != nil {
-		t.Fatal(err)
-	}
-	if !ok {
-		t.Fatal("not found")
-		return
-	}
-	t.Logf("Latest %s %s", mod.Path, mod.MaxVersion("", false))
-}
-
-func TestQueryPackage(t *testing.T) {
-	tests := []struct {
-		pkgpath string
-		modpath string
-	}{
-		{
-			pkgpath: "github.com/go-redis/redis",
-			modpath: "github.com/go-redis/redis",
-		},
-		{
-			pkgpath: "github.com/google/go-cmp/cmp",
-			modpath: "github.com/google/go-cmp",
-		},
-		{
-			pkgpath: "github.com/go-git/go-git/v5/plumbing/format/commitgraph",
-			modpath: "github.com/go-git/go-git/v5",
-		},
-	}
-	for _, tt := range tests {
-		t.Run(tt.pkgpath, func(t *testing.T) {
-			mod, err := QueryPackage(tt.pkgpath, true)
-			if err != nil {
-				t.Fatal(err)
-			}
-			if mod.Path != tt.modpath {
-				t.Fatalf("invalid modpath, want %q, got %q", tt.modpath, mod.Path)
-			}
-		})
-	}
-}
-
 func TestModule(t *testing.T) {
 	tests := []struct {
 		mod      *Module
@@ -189,30 +133,6 @@ func TestModule(t *testing.T) {
 	}
 }
 
-func TestMaxVersion(t *testing.T) {
-	tests := []struct {
-		lo, hi string
-	}{
-		{"v0.0.0", "v0.0.1"},
-		{"v0.2.0", "v1.0.0"},
-		{"v3.0.0+incompatible", "v0.0.1"},
-		{"v3.0.0+incompatible", "v5.0.1+incompatible"},
-		{"", "v6.14.1+incompatible"},
-		{"invalid", ""},
-	}
-	for _, tt := range tests {
-		name := fmt.Sprintf("%s < %s", tt.lo, tt.hi)
-		t.Run(name, func(t *testing.T) {
-			if got := MaxVersion(tt.lo, tt.hi); got != tt.hi {
-				t.Fatalf("MaxVersion(%q, %q) = %q", tt.lo, tt.hi, got)
-			}
-			if got := MaxVersion(tt.hi, tt.lo); got != tt.hi {
-				t.Fatalf("MaxVersion(%q, %q) = %q", tt.hi, tt.lo, got)
-			}
-		})
-	}
-}
-
 func TestIsNewerVersion(t *testing.T) {
 	tests := []struct {
 		old, new     string
@@ -249,3 +169,219 @@ func TestIsNewerVersion(t *testing.T) {
 		})
 	}
 }
+
+func TestCompareVersion(t *testing.T) {
+	tests := []struct {
+		v, w string
+		want int
+	}{
+		{v: "v0.0.0", w: "v1.0.0", want: -1},
+		{v: "v1.0.0", w: "v0.0.0", want: 1},
+		{v: "v0.0.0", w: "v0.0.0", want: 0},
+		{v: "v12.0.0+incompatible", w: "v0.0.0", want: -1},
+		{v: "", w: "", want: 0},
+		{v: "v0.1.0", w: "bad", want: 1},
+		{v: "v0.0.0+incompatible", w: "v0.0.0", want: -1},
+		{v: "v0.0.0", w: "v0.0.1", want: -1},
+		{v: "v0.2.0", w: "v1.0.0", want: -1},
+		{v: "v3.0.0+incompatible", w: "v0.0.1", want: -1},
+		{v: "v3.0.0+incompatible", w: "v5.0.1+incompatible", want: -1},
+		{v: "", w: "v6.14.1+incompatible", want: -1},
+		{v: "invalid", w: ""},
+	}
+	for _, tt := range tests {
+		t.Run("", func(t *testing.T) {
+			if got := CompareVersion(tt.v, tt.w); got != tt.want {
+				t.Fatalf("CompareVersion(%q, %q) = %v, want %v", tt.v, tt.w, got, tt.want)
+			}
+		})
+	}
+}
+
+func TestVersionRange(t *testing.T) {
+	tests := []struct {
+		r    VersionRange
+		v    string
+		want bool
+	}{
+		{
+			r:    VersionRange{Low: "v0.0.0", High: "v0.0.1"},
+			v:    "v0.0.0",
+			want: true,
+		},
+		{
+			r:    VersionRange{Low: "v0.0.0", High: "v0.0.1"},
+			v:    "v0.0.1",
+			want: true,
+		},
+		{
+			r:    VersionRange{Low: "v0.0.0", High: "v0.0.1"},
+			v:    "v0.0.2",
+			want: false,
+		},
+		{
+			r:    VersionRange{Low: "v0.0.0", High: "v0.0.1"},
+			v:    "v0.0.0+incompatible",
+			want: false,
+		},
+	}
+	for _, tt := range tests {
+		t.Run("", func(t *testing.T) {
+			if got := tt.r.Includes(tt.v); got != tt.want {
+				t.Fatalf("VersionRange{Low: %q, High: %q}.Includes(%q) = %v, want %v", tt.r.Low, tt.r.High, tt.v, got, tt.want)
+			}
+		})
+	}
+}
+
+func TestQuery(t *testing.T) {
+	// Start the test mod proxy
+	proxy, err := testmodproxy.Load("testdata/modules")
+	if err != nil {
+		t.Fatal(err)
+	}
+	server := httptest.NewServer(proxy)
+	defer server.Close()
+	t.Setenv("GOPROXY", server.URL)
+
+	tests := []struct {
+		name    string
+		modpath string
+		want    *Module
+		exist   bool
+	}{
+		{
+			name:    "existing module",
+			modpath: "example.com/testmod",
+			want: &Module{
+				Path:     "example.com/testmod",
+				Versions: []string{"v1.0.0", "v1.1.0", "v1.2.0"},
+			},
+			exist: true,
+		},
+		{
+			name:    "non-existent module",
+			modpath: "example.com/nonexistent",
+			exist:   false,
+		},
+	}
+
+	for _, tt := range tests {
+		t.Run(tt.name, func(t *testing.T) {
+			mod, ok, err := Query(tt.modpath, false)
+			if err != nil {
+				t.Fatal(err)
+			}
+			if ok != tt.exist {
+				t.Fatalf("Query() ok = %v, want %v", ok, tt.exist)
+			}
+			if !reflect.DeepEqual(mod, tt.want) {
+				t.Fatalf("Query() = %+v, want %+v", mod, tt.want)
+			}
+		})
+	}
+}
+
+func TestLatest(t *testing.T) {
+	// Start the test mod proxy
+	proxy, err := testmodproxy.Load("testdata/modules")
+	if err != nil {
+		t.Fatal(err)
+	}
+	server := httptest.NewServer(proxy)
+	defer server.Close()
+	t.Setenv("GOPROXY", server.URL)
+
+	tests := []struct {
+		name    string
+		modpath string
+		pre     bool
+		want    *Module
+	}{
+		{
+			name:    "latest from v1 base",
+			modpath: "example.com/testmod",
+			pre:     false,
+			want: &Module{
+				Path:     "example.com/testmod/v3",
+				Versions: []string{"v3.0.0"},
+			},
+		},
+		{
+			name:    "latest from v2 path",
+			modpath: "example.com/testmod/v2",
+			pre:     false,
+			want: &Module{
+				Path:     "example.com/testmod/v3",
+				Versions: []string{"v3.0.0"},
+			},
+		},
+		{
+			name:    "latest from v3 path",
+			modpath: "example.com/testmod/v3",
+			pre:     false,
+			want: &Module{
+				Path:     "example.com/testmod/v3",
+				Versions: []string{"v3.0.0"},
+			},
+		},
+	}
+
+	for _, tt := range tests {
+		t.Run(tt.name, func(t *testing.T) {
+			mod, err := Latest(tt.modpath, false, tt.pre)
+			if err != nil {
+				t.Fatal(err)
+			}
+			if !reflect.DeepEqual(mod, tt.want) {
+				t.Fatalf("Latest() = %+v, want %+v", mod, tt.want)
+			}
+		})
+	}
+}
+
+func TestQueryPackage(t *testing.T) {
+	// Start the test mod proxy
+	proxy, err := testmodproxy.Load("testdata/modules")
+	if err != nil {
+		t.Fatal(err)
+	}
+	server := httptest.NewServer(proxy)
+	defer server.Close()
+	t.Setenv("GOPROXY", server.URL)
+
+	tests := []struct {
+		name    string
+		pkgpath string
+		want    *Module
+	}{
+		{
+			name:    "module root",
+			pkgpath: "example.com/testmod",
+			want: &Module{
+				Path:     "example.com/testmod",
+				Versions: []string{"v1.0.0", "v1.1.0", "v1.2.0"},
+			},
+		},
+		{
+			name:    "existing module package",
+			pkgpath: "example.com/testmod/v2/pkg",
+			want: &Module{
+				Path:     "example.com/testmod/v2",
+				Versions: []string{"v2.0.0", "v2.1.0"},
+			},
+		},
+	}
+
+	for _, tt := range tests {
+		t.Run(tt.name, func(t *testing.T) {
+			mod, err := QueryPackage(tt.pkgpath, false)
+			if err != nil {
+				t.Fatal(err)
+			}
+			if !reflect.DeepEqual(mod, tt.want) {
+				t.Fatalf("QueryPackage() = %+v, want %+v", mod, tt.want)
+			}
+		})
+	}
+}
